@@ -88,7 +88,7 @@
 	(,@valid-part))))
 
 (defun mucs-ccl-if-invalid-repeat ()
-  (mucs-ccl-if-invalid '((loop))))
+  (mucs-ccl-if-invalid '((repeat))))
 
 ;;;
 ;;; Common I/O interface.
@@ -113,54 +113,23 @@
   `((read r0)))
 
 (defvar mucs-ccl-read-ex-be-2-octet
-  `((read r0 r6)
-    (r0 = (r0 <8 r6))))
+  `((read r0 r4)
+    (r0 = (r0 <8 r4))))
 
 (defvar mucs-ccl-read-ex-le-2-octet
-  `((read r0 r6)
-    (r0 = (r6 <8 r0))))
-
-(defvar mucs-ccl-read-ex-be-4-octet
-  `((read r0 r6)
-    (r0 = (r0 <8 r6))
-    (read r4 r6)
-    (r0 |= (r4 << 16))))
-
-(defvar mucs-ccl-read-ex-le-4-octet
-  `((read r0 r6)
-    (r0 |= (r6 << 16))
-    (read r4 r6)
-    (r0 |= (r4 <8 r6))))
+  `((read r0 r4)
+    (r0 = (r4 <8 r0))))
 
 (defvar mucs-ccl-write-ex-1-octet
   `((write r0)))
 
 (defvar mucs-ccl-write-ex-be-2-octet
-  `((r6 = (r0 >> 8))
-    (r4 = (r0 & ?\xff))
-    (write r6 r4)))
+  `((write (r0 >> 8))
+    (write (r0 & ?\xff))))
 
 (defvar mucs-ccl-write-ex-le-2-octet
-  `((r6 = (r0 >> 8))
-    (r4 = (r0 & ?\xff))
-    (write r4 r6)))
-
-(defvar mucs-ccl-write-ex-be-4-octet
-  `((r4 = ((r0 >> 16) & ?\xff))
-    (r6 = 0)
-    (write r6 r4)
-    (r6 = ((r0 >> 8) & ?\xff))
-    (r4 = (r0 & ?\xff))    
-    (write r6 r4)))
-
-(defvar mucs-ccl-write-ex-le-4-octet
-  `((r4 = (r0 & ?\xff))
-    (r6 = ((r0 >> 8) & ?\xff))
-    (write r4 r6)
-    (r4 = ((r0 >> 16) & ?\xff))
-    (r6 = 0)
-    (write r4 r6)))
-
+  `((write (r0 & ?\xff))
+    (write (r0 >> 8))))
 
 ;;;
 ;;; generic type definition
@@ -202,41 +171,73 @@
  'char-packed-1
  'identity)
 
+;;
+;; Internal state management functions
+;;
+(defvar mucs-ccl-internal-state-usage
+  (make-vector mucs-code-range-bits nil))
+
+(defun mucs-ccl-internal-state-reserve (name bit)
+  (if (not (symbolp name))
+      (signal 'wrong-type-argument
+	      name))
+  (if (>= bit (length mucs-ccl-internal-state-usage))
+      (error "Bit %d is too large.(%s)" bit name))
+  (let ((slot (aref mucs-ccl-internal-state-usage bit)))
+    (if (and name slot
+	     (not (eq name slot)))
+	(error "Bit %d have been already reserved.(%s)" bit name))
+    (aset mucs-ccl-internal-state-usage bit name)
+    (if name
+	(put name 'mucs-ccl-internal-state-bit bit)
+      (if slot
+	  (put slot
+	       'mucs-ccl-internal-state-bit nil)))))
+
+(defun mucs-ccl-set-internal-state (name state)
+  "Return a block that sets internal state to STATE for NAME."
+  (let ((bit (get name 'mucs-ccl-internal-state-bit)))
+    (if (not bit)
+	(error "%s is not registered for internal state!"
+	       name))
+    (if state
+	`((r5 |= ,(lsh 1 bit)))
+      `((r5 &= ,(logxor (mucs-number-mask) (lsh 1 bit)))))))
+
+(defun mucs-ccl-check-internal-state (name)
+  "Return a statement that checks internal state for NAME."
+  `(r5 & ,(lsh 1
+	       (or (get name 'mucs-ccl-internal-state-bit)
+		   (error "%s is not registered for internal state!"
+			  name)))))
 
 ;;
 ;; Dealing with line separator problem.
 ;;
 
-(defconst mucs-ccl-previous-cr-p-bit 16)
-
-(defun mucs-ccl-previous-cr-p ()
-  '(r5 & ?\x2))
-
-(defun mucs-ccl-set-eol-state (eol)
-  (if (eq eol 'cr)
-      '((r5 |= ?\x2))
-    `((r5 &= ,(logxor (mucs-number-mask) ?\x2)))))
+(mucs-ccl-internal-state-reserve 'previous-cr-p 0)
 
 (defun mucs-ccl-write-char-packed-1-dos ()
   `((if (r0 == ?\x0d)
-	((if ,(mucs-ccl-previous-cr-p)
+	((if ,(mucs-ccl-check-internal-state 'previous-cr-p)
 	     ((write ?\x0d))
-	   ,(mucs-ccl-set-eol-state 'cr)))
-      ((r6 = (r0 != ?\x0a))
-       (if (,(mucs-ccl-previous-cr-p) & r6)
+	   ,(mucs-ccl-set-internal-state 'previous-cr-p t)))
+      ((r4 = (r0 != ?\x0a))
+       (if (,(mucs-ccl-check-internal-state 'previous-cr-p) & r4)
 	   ((write ?\x0d)))
-       ,@(mucs-ccl-set-eol-state nil)
-       ,@(mucs-ccl-write-char-packed-1)))))
+       (,@(mucs-ccl-set-internal-state 'previous-cr-p nil)
+	,@(mucs-ccl-write-char-packed-1))))))
 
 (defun mucs-ccl-write-char-1-dos ()
   `((if (r0 == ?\x0d)
-	((if ,(mucs-ccl-previous-cr-p)
+	((if ,(mucs-ccl-check-internal-state 'previous-cr-p)
 	     ((write ?\x0d))
-	   ,(mucs-ccl-set-eol-state 'cr)))
-      ((r6 = (r0 != ?\x0a))
-       (if (,(mucs-ccl-previous-cr-p) & r6)
+	   ,(mucs-ccl-set-internal-state 'previous-cr-p t)))
+      ((r4 = (r0 != ?\x0a))
+       (if (,(mucs-ccl-check-internal-state 'previous-cr-p)
+	    & r4)
 	   ((write ?\x0d)))
-       ,@(mucs-ccl-set-eol-state nil)
+       ,@(mucs-ccl-set-internal-state 'previous-cr-p nil)
        (write r0)))))
 
 (mucs-type-register-serialization
